package ims.headfinder;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.zip.ZipOutputStream;

import ims.data.Instance;
import ims.hotcoref.symbols.SymbolTable;
import ims.io.Format;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.io.SentenceWriter;
import ims.perceptron.FV3;
import ims.perceptron.IFV;
import ims.perceptron.Long2IntFactory;
import ims.perceptron.Long2IntInterface;
import ims.perceptron.ParametersFloat;
import ims.perceptron.ParametersFloat.IUpdStruct;
import ims.trans.InstanceMapping;
import ims.trans.LossFunction;
import ims.trans.Options;
import ims.trans.labelfilter.ILabelFilter;
import ims.trans.labelfilter.LabelFilterFactory;
import ims.trans.util.Eval;
import ims.util.Accuracy;
import ims.util.ArrayFunctions;
import ims.util.ArrayShuffler;
import ims.util.DBO;
import ims.util.Pair;
import ims.util.ZipUtil;

public class Train {

	public static void main(String[] args) throws IOException, ClassNotFoundException{
		Options options=new Options(args);
		options.sanityCheckTestEvery();
		options.warnNoModel();
		
		
		SentenceReader<?> reader = ReaderWriterFactory.getReader(options.train, options.format);
		
		SymbolTable symTab=new SymbolTable(options.useStag,false,options.useShape);
		InstanceMapping instanceMapping=new InstanceMapping(options.useStag,options.useStack,options.useShape);
		
		List<Instance> instances=createInstances(options, reader, symTab,	instanceMapping);

		symTab.freezeMappings();
		
		if(Options.shuffleStrategy.shuffleFirstIteration){
			DBO.println("Shuffling instances");
			new ArrayShuffler().shuffle(instances);
		}
		
		Long2IntInterface l2i=Long2IntFactory.getLong2Int(options);
		ParametersFloat parameters=ParametersFloat.createParameters(options);
		ILabelFilter labels=LabelFilterFactory.getLabelFilter(symTab, instances, options.labelFilter);
		
		Decoder decoder=new Decoder(parameters.parameters, l2i, symTab, instanceMapping,labels);
		LossFunction lossFunction=LossFunction.getLossFunction(Options.lossFunctionType);

		DBO.println("Now begin training");
		
		Model m=train(instances,symTab,l2i,parameters,decoder,lossFunction,options);
		
		if(options.model!=null){
			DBO.println("Saving model...");
			ZipOutputStream zos=ZipUtil.openZOS(options.model);
			ZipUtil.writeObjectAsZipEntry(zos, Model.ZIP_ENTRY, m);
			zos.close();
			DBO.println("Done saving model.");
		}
		
		options.done();
	}

	private static Model train(List<Instance> instances, SymbolTable symTab, Long2IntInterface l2i, ParametersFloat parameters, Decoder decoder, LossFunction lossFunction, Options options) throws IOException {
		int itemCount=0;
		for(int iter=0;iter<options.iterations;++iter){
			DBO.println("Iteration: "+iter);
			long start=System.currentTimeMillis();
			IFV<FV3> goldFV=new FV3();
			IFV<FV3> predFV=new FV3();
			int erase=0;
			int correct=0;
			int isCount=0;
			DBO.printNoPrefix("Sent: ");
			for(Instance is:instances){
				++isCount;
				++itemCount;
				if(isCount%111==0)
					erase=DBO.eraseAndPrint(erase, infoString(isCount,correct,start));
				
				Pair<int[],int[]> p=decoder.parse(is); 
				
				int[] pHeads=p.getLeft();
				int[] pLbls=p.getRight();

				goldFV.clear();
				predFV.clear();
				double loss=0.d;
				
				for(int dep=1;dep<pHeads.length;++dep){
					boolean isCorrect=pHeads[dep]==is.head[dep] && pLbls[dep]==is.lbl[dep];
					if(isCorrect)
						continue;
					else 
						loss+=extractFeaturesToFVs(is,dep,pHeads[dep],pLbls[dep],goldFV,predFV,decoder);
					
				}
				
				//Ok, prediction is done. Update if loss>0
				if(loss>0){
					doUpdate(goldFV,predFV,parameters,loss,itemCount);
				} else {
					++correct;
				}
			}
			DBO.eraseAndPrint(erase, infoString(isCount,correct,start));
			DBO.println();
			if(options.testEveryIter && options.test!=null && options.gold!=null && options.out!=null){
				float[] params;
				if(Options.useAveraging)
					params=parameters.averagedCopy(itemCount);
				else
					params=parameters.parameters;
				Decoder d=new Decoder(params, l2i, symTab, decoder.instanceMapping,decoder.labels);
				interIterTest(d,options.test,options.out,options.gold,options.format);
			}
			if(Options.shuffleStrategy.shuffleBetweenIterations)
				ArrayShuffler.SINGLETON.shuffle(instances);
		}
		
		if(Options.useAveraging)
			parameters.average(itemCount);

		return new Model(parameters.parameters, l2i, symTab, decoder.instanceMapping,decoder.labels);
	}

	private static void interIterTest(Decoder d, File test, File out, File gold, Format format) throws IOException {
		SentenceReader<?> reader=ReaderWriterFactory.getReader(test, format);
		SentenceWriter writer=ReaderWriterFactory.getWriter(out);
		Test.test(reader, writer, d);
		Pair<Accuracy,Accuracy> a=Eval.getUASandLAS(out, gold, format);
		System.out.println("UAS: "+a.getLeft().toString() + "   LAS: "+a.getRight().toString());
	}

	private static String infoString(int isCount, int correct, long start) {
		long sps=(System.currentTimeMillis()-start)/isCount;
		return String.format("%-5d   c: %-5d   | %d/s", isCount,correct,sps);
	}

	private static void doUpdate(IFV<FV3> goldFV, IFV<FV3> predFV, ParametersFloat parameters, double loss,int itemCount) {
		double goldScore=goldFV.getScore(parameters.parameters);
		double predScore=predFV.getScore(parameters.parameters);
		
//		PerceptronUpdStruct upd=new PerceptronUpdStruct(predScore,goldScore,loss,itemCount,goldFV.getDistVector(predFV));
		IUpdStruct upd=parameters.getUpdStruct(predScore,goldScore,loss,itemCount,goldFV.getDistVector(predFV));
		parameters.update(upd);
	}

	static final long[] sink=new long[FE.SINK_LEN];
	private static double extractFeaturesToFVs(Instance is,int dep,int bestHead,int bestLabel,IFV<FV3> goldFV,IFV<FV3> predFV,Decoder decoder){
		boolean correctHead=is.head[dep]==bestHead;
		if(correctHead){
			int features=decoder.fe.extractFeatures(is, bestHead, dep, sink);
			ArrayFunctions.shiftAllLeftByBits(sink, decoder.lblBits, features);
			for(int i=0;i<features;++i){
				goldFV.add(decoder.l2i.l2i(sink[i]|is.lbl[dep]));
				predFV.add(decoder.l2i.l2i(sink[i]|bestLabel));
			}
			return 0.5d;
		} else {
			//First gold
			{
				int features=decoder.fe.extractFeatures(is, is.head[dep], dep, sink);
				ArrayFunctions.shiftAllLeftByBits(sink, decoder.lblBits, features);
				for(int i=0;i<features;++i)
					goldFV.add(decoder.l2i.l2i(sink[i]|is.lbl[dep]));
			}
			//Then pred
			{
				int features=decoder.fe.extractFeatures(is, bestHead, dep, sink);
				ArrayFunctions.shiftAllLeftByBits(sink, decoder.lblBits, features);
				for(int i=0;i<features;++i)
					predFV.add(decoder.l2i.l2i(sink[i]|bestLabel));
			}
			return 1.0d;
		}
	}
	
	private static List<Instance> createInstances(Options options, SentenceReader<?> reader, SymbolTable symTab, InstanceMapping instanceMapping) {
		int instCount=symTab.registerSimple(reader, options.count, options.labeled);
		List<Instance> instances=instanceMapping.createTrainInstances(reader,symTab,instCount,null,options.labeled);
		return instances;
	}
	
}
