package ims.tag;

import java.io.Serializable;
import java.util.Random;

import ims.hotcoref.symbols.TagSymbolTable;
import ims.perceptron.FV3;
import ims.perceptron.IFV;
import ims.tag.features.AbstractFeatureSet;
import ims.tag.features.FeatureScorer;
import ims.tag.features.Potentials;
import ims.util.ArrayFunctions;
import ims.util.ProbSampler;

public class SamplingDecoder implements IDecoder {
	private static final long serialVersionUID = 1581882504625214737L;
	
	private final FeatureScorer featureScorer;
	private final ProbSampler probSampler;
	private final Random random=new Random(234562342l);
	
	private final int greedyRestarts=TagOptions.greedyRestarts;
	private final int randomRestarts=TagOptions.randomRestarts;
	private final int resamplings=TagOptions.resamplings;
	private final int tagChanges=TagOptions.tagChanges;
	private final boolean entropyPropose=TagOptions.entropyPropose;
	
	private final IAcceptor acceptor;
	
	public SamplingDecoder(FeatureScorer featureScorer){
		this.featureScorer=featureScorer;
		this.probSampler=new ProbSampler(235292990l);
		this.acceptor=getAcceptor(TagOptions.sampleAcceptor);
	}
	
	@Override
	public int[] decodeInstance(TagInstance instance, TagSymbolTable symTab) {
		
		Potentials potentials=featureScorer.getPotentials(instance);
		int[] greedyAssignment=GreedyLocalDecoder.greedyFromPotentials(instance, potentials);
		double greedyScore=scoreTagging(potentials,instance,new SamplingTagContext(greedyAssignment));
		
		
		int[] bestTagging=greedyAssignment;
		double bestScore=greedyScore;
		
		double[] entropies=getEntropies(potentials);
		
		//Do restarts from the best tagging
		for(int r=0;r<greedyRestarts;++r){
			SamplingTagContext tagContext=new SamplingTagContext(greedyAssignment);
			double tagsScore=greedyScore;
			for(int i=0;i<resamplings;++i){
				int changes=resample(tagContext,potentials,entropies);
				if(changes>0){
					double score=scoreTagging(potentials,instance,tagContext);
					if(acceptor.acceptChange(tagsScore, score)){
						tagContext.commitModifications();
						tagsScore=score;
					} else {
						tagContext.rollbackModifications();
					}
				}
			}
			if(tagsScore>bestScore){
				bestTagging=tagContext.getTags();
				bestScore=tagsScore;
			}
		}

		//Do restarts from random initialization
		for(int r=0;r<randomRestarts;++r){
			SamplingTagContext tagContext=sampleTagging(potentials,instance);
			double tagsScore=scoreTagging(potentials,instance,tagContext);
			for(int i=0;i<resamplings;++i){
				int changes=resample(tagContext,potentials,entropies);
				if(changes>0){
					double score=scoreTagging(potentials,instance,tagContext);
					if(acceptor.acceptChange(tagsScore, score)){
						tagContext.commitModifications();
						tagsScore=score;
					} else {
						tagContext.rollbackModifications();
					}
				}
			}
			if(tagsScore>bestScore){
				bestTagging=tagContext.getTags();
				bestScore=tagsScore;
			}
		}
		
		return bestTagging;
	}

	private double[] getEntropies(Potentials potentials) {
		double[] e=new double[potentials.getLength()];
		for(int i=1;i<e.length;++i)
			e[i]=getEntropy(potentials.getPotentialArray(i));
		return e;
	}
	
	public static double getEntropy(double[] potentials){
		double r=0;
		double[] probabilities=new double[potentials.length];
		double expSum=0;
		for(int i=2;i<potentials.length;++i){
			double exped=Math.exp(potentials[i]);
			probabilities[i]=exped;
			expSum+=exped;
		}
		for(int i=2;i<potentials.length;++i){
			probabilities[i]/=expSum;
		}
		for(int i=2;i<probabilities.length;++i){
			double d=probabilities[i];
			r-=(d*Math.log(d));
		}
		return r;
	}

	private int resample(SamplingTagContext tagContext, Potentials potentials, double[] entropy) {
		int changes=0;
		for(int i=0;i<tagChanges;++i){
			int idx=proposeIndex(tagContext.length(),entropy);
			int t=probSampler.sampleX(potentials.getPotentialArray(idx));
			if(tagContext.getTag(idx)!=t){
				++changes;
				tagContext.modify(idx, t);
			}
		}
		return changes; //XXX changes might be 2 if the same index was changed twice. So it's really just reliable as a binary indicator whether changes were made or not. Change the logic to really count number of token changes if necessary.
	}

	private int proposeIndex(int length,double[] entropies) {
		final int i;
		if(entropyPropose){
			double sum=ArrayFunctions.sum(entropies);
			i=ProbSampler.rouletteSampleEntry(random,entropies,sum);
		} else {
			i=random.nextInt(length-1)+1;
		}
		return i;
	}
	
	private double scoreTagging(Potentials potentials,TagInstance inst,SamplingTagContext tagContext) {
		double score=0.d;
		for(int i=1;i<tagContext.length();++i){
			double local=potentials.getPotential(i, tagContext.getTag(i));
			double nonlocal=featureScorer.scoreNonLocal(i, inst, tagContext, tagContext.getTag(i));
			score+=local+nonlocal;
		}
		return score;
	}

	private SamplingTagContext sampleTagging(Potentials potentials,TagInstance inst) {
		int[] r=new int[inst.forms.length];
		for(int i=1;i<inst.forms.length;++i){
			r[i]=probSampler.sampleX(potentials.getPotentialArray(i));
		}
		return new SamplingTagContext(r); //XXX can be improved, here's a copy operation in the constructor
	}

	@Override
	public double recoverFeaturesAndFillFV(int[] tags, TagInstance inst, AbstractFeatureSet fs, IFV<FV3> fv, FeatureScorer featureScorer) {
		TagContext tagContext=new TagContext(tags);
		fv.clear();
		double score=0.d;
		for(int i=1;i<tags.length;++i){
			score+=fs.recoverFeaturesAndFillFV(inst, tagContext, i, tags[i], featureScorer, fv);
		}
		return score;
	}

	
	
	
	public static interface IAcceptor extends Serializable {
		public enum Type { Always, OnlyIncrease, CoinFlip }
		public boolean acceptChange(double oldScore,double newScore);
	}
	
	public static IAcceptor getAcceptor(IAcceptor.Type type){
		switch(type){
		case Always:       return new AlwaysAcceptor();
		case OnlyIncrease: return new OnlyIncreaseAcceptor();
		case CoinFlip:     return new CoinFlipAcceptor(TagOptions.coinFlipAcceptorP,-99024712451l);
		default:           throw new Error("not implemneted");
		}
	}
	
	static class AlwaysAcceptor implements IAcceptor {
		private static final long serialVersionUID = -5623175044264545335L;
		@Override
		public boolean acceptChange(double oldScore, double newScore) {
			return true;
		}
	}
	
	static class OnlyIncreaseAcceptor implements IAcceptor {
		private static final long serialVersionUID = 4044733244776111595L;
		@Override
		public boolean acceptChange(double oldScore, double newScore) {
			return newScore>oldScore;
		}
	}
	
	static class CoinFlipAcceptor implements IAcceptor {
		private static final long serialVersionUID = -8597477578854980705L;
		private final double p;
		private final Random r;
		
		public CoinFlipAcceptor(double p,long randSeed){
			this.p=p; //Chance of accepting change
			this.r=new Random(randSeed);
		}
		
		@Override
		public boolean acceptChange(double oldScore, double newScore) {
			double d=r.nextDouble();
			if(d<p)
				return true; //keep if d<p. Means p is the probability of accepting the change.
			else
				return false;
		}
		
	}
}
