package ims.tag;

import java.util.Arrays;

import ims.hotcoref.symbols.AbstractSymbolMapping;
import ims.hotcoref.symbols.TagSymbolTable;
import ims.perceptron.FV3;
import ims.perceptron.IFV;
import ims.tag.features.AbstractFeatureSet;
import ims.tag.features.FeatureScorer;
import ims.tag.features.Potentials;

public class SecondOrderViterbiDecoder implements IDecoder {
	private static final long serialVersionUID = -1835328311021529398L;

	static final int NOTAG=AbstractSymbolMapping.NONE_IDX;

	private static final boolean FAST=true;
	
	private final FeatureScorer featureScorer;
	
	public SecondOrderViterbiDecoder(FeatureScorer featureScorer){
		this.featureScorer=featureScorer;
	}
	
	@Override
	public int[] decodeInstance(TagInstance instance, TagSymbolTable symTab) {
		//tri-grams: w,u,v -- v is current, u is prev, and w is two behind
		
		Potentials potentials=featureScorer.getPotentials(instance);
		
		int tagMax=symTab.tags.getItems();
		
		TableItem[][][] table=new TableItem[instance.forms.length][tagMax][tagMax]; //indexed by (position,tag u,tag v)
		
		TagContext context=new TagContext(instance.forms.length);
		
		//Do the initialization
		{ //Init for v=1
			int v=1;
			int uT=NOTAG;
			int wT=NOTAG;
			double[] trans=featureScorer.scoreNonLocal(v, instance, context);
			for(int vT:symTab.tagIdx){
				double potentialScore=potentials.getPotential(v, vT);
//				double tailScore=0.d; //we don't need this since it's always zero 
				double transScore=trans[vT];
				double score=potentialScore+transScore;
				table[1][NOTAG][vT]=new TableItem(null,score,vT);
			}
		}
		if(instance.forms.length>2){
			{ //Init for v=2
				int v=2;
				int wT=NOTAG;
				for(int uT:symTab.tagIdx){
					context.setTag(v-1, uT);
					double[] trans=featureScorer.scoreNonLocal(v, instance, context);
					for(int vT:symTab.tagIdx){
						double potentialScore=potentials.getPotential(v, vT);
						double transScore=trans[vT];
						double tailScore=table[v-1][wT][uT].totalScore;
						double score=potentialScore+transScore+tailScore;
						TableItem prev=table[v-1][wT][uT];
						TableItem cur=new TableItem(prev,score,vT);
						table[v][uT][vT]=cur;
					}
				}
			}
			if(!FAST){
				for(int v=3;v<instance.forms.length;++v){
					//loop over tags
					for(int vT:symTab.tagIdx){
						for(int uT:symTab.tagIdx){
							double bestScore=-99999.d;
							int bestW=-1;
							for(int wT:symTab.tagIdx){
								context.setTag(v-1, uT);
								context.setTag(v-2, wT);
								double[] trans=featureScorer.scoreNonLocal(v, instance, context); //XXX this is very inefficient -- should rearrange the loops and have for(u){for(w){for(v){}}}, and add extra bookkeeping and move the scoring out 
								double potentialScore=potentials.getPotential(v, vT);
								double transScore=trans[vT];
								double tailScore=table[v-1][wT][uT].totalScore;
								double score=potentialScore+transScore+tailScore;
								if(score>bestScore){
									bestScore=score;
									bestW=wT;
								}
							}
							TableItem prev=table[v-1][bestW][uT]; //Best way to get here
							TableItem cur=new TableItem(prev,bestScore,vT);
							table[v][uT][vT]=cur;
						}
					}
				}
			} else {
				for(int v=3;v<instance.forms.length;++v){
					for(int uT:symTab.tagIdx){
						context.setTag(v-1, uT);
						int[]    bestWTbyVT=new int[symTab.tags.getItems()];             //This is the best wT for each vT
						double[] bestTransAndTailScoreByWT=new double[symTab.tags.getItems()];  //This is the best transition score for the vT given the wT in the above array
						Arrays.fill(bestTransAndTailScoreByWT, -99999.d);
						for(int wT:symTab.tagIdx){
							context.setTag(v-2, wT);
							double[] transScore=featureScorer.scoreNonLocal(v, instance, context);
							for(int vT:symTab.tagIdx){
								if(transScore[vT]+table[v-1][wT][uT].totalScore>bestTransAndTailScoreByWT[vT]){
									bestTransAndTailScoreByWT[vT]=transScore[vT]+table[v-1][wT][uT].totalScore;
									bestWTbyVT[vT]=wT;
								}
							}
						}
						for(int vT:symTab.tagIdx){
							int wT=bestWTbyVT[vT];
							double transAndTailScore=bestTransAndTailScoreByWT[vT];
							double potentialScore=potentials.getPotential(v, vT);
							TableItem tail=table[v-1][wT][uT];
							double score=transAndTailScore+potentialScore;
							TableItem cur=new TableItem(tail,score,vT);
							table[v][uT][vT]=cur;
						}
					}
				}
			}
		}
		//Unravel backpointers
		
		//Find best
		TableItem best=null;
		for(int u=0;u<tagMax;++u){
			for(int v=0;v<tagMax;++v){
				TableItem ti=table[instance.forms.length-1][u][v];
				if(ti==null)
					continue;
				if(best==null || ti.totalScore>best.totalScore)
					best=ti;
			}
		}
		//Follow backpointers
		int[] tags=new int[instance.forms.length];
		for(int c=tags.length-1;c>0;--c){
			tags[c]=best.tag;
			best=best.prev;
		}
		return tags;
	}

	static class TableItem {

		double totalScore;
		int tag;
		TableItem prev;
		
		public TableItem(TableItem prev, double bestScore, int tag) {
			this.totalScore=bestScore;
			this.tag=tag;
			this.prev=prev;
		}
		
	}

	@Override
	public double recoverFeaturesAndFillFV(int[] tags, TagInstance inst, AbstractFeatureSet fs, IFV<FV3> fv, FeatureScorer featureScorer) {
		TagContext tagContext=new TagContext(tags.length);
		fv.clear();
		double score=0.d;
		for(int i=1;i<tags.length;++i){
			score+=fs.recoverFeaturesAndFillFV(inst, tagContext, i, tags[i], featureScorer, fv);
			tagContext.setTag(i, tags[i]);
		}
		return score;
	}
}
