package ims.trans;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import ims.data.ISentence;
import ims.data.Instance;
import ims.data.Parse;
import ims.hotcoref.symbols.SymbolTable;
import ims.oracle.IOracle;
import ims.oracle.Configuration.TR;
import ims.perceptron.Long2IntInterface;
import ims.trans.transitions.ITransition;
import ims.trans.transitions.ITransitionFilter;
import ims.trans.transitions.TransitionFactory.LabeledAbstractTransition;
import ims.util.ArrayFunctions;
import ims.util.IntPair;
import ims.util.MinHeap;
import ims.util.Pair;

public class Decoder extends AbstractDecoder {
	
	private int beam;
	private int goldBeam;
	
	public Decoder(Model m){
		this(m.parameters,m.l2i,m.symTab,m.transitionFilter,m.fe,m.mergeStates,m.labeled,m.instanceMapping,m.beamSize,-1);
	}
	
	public Decoder(float[] parameters, Long2IntInterface l2i, SymbolTable symTab,ITransitionFilter transitionFilter,FeatureExtractor featureExtractor,boolean mergeStates,boolean labeled,InstanceMapping instanceMapping,int beam,int goldBeam) {
		super(parameters, l2i, symTab, transitionFilter, featureExtractor, mergeStates, labeled, instanceMapping);
		this.beam=beam;
		this.goldBeam=goldBeam;
	}
	
	static class ExpansionItem extends MinHeap.AbstractTieBreakingHeapItem<ExpansionItem> {
		//right now this is the total score. These things are comparable as long all input states that are being expanded have the same number of transitions
		final double newTotalScore;
		final ITransition trans;
		final State fromState;
		
		public ExpansionItem(ITransition trans, double newTotalScore, State fromState,int tieBreaker) {
			super(tieBreaker);
			this.trans = trans;
			this.newTotalScore = newTotalScore;
			this.fromState = fromState;
		}

		@Override
		public double score() {
			return newTotalScore;
		}
		
		public State execute(){
			State r=trans.apply(fromState, newTotalScore);
			return r;
		}

		public String toString(){
			return trans.getFullTransitionLabel()+"\nScore: "+newTotalScore;
		}
	}
	
	public interface ITROracle {
		public List<ITransition> getCorrectTransitions(State s,Instance inst,ITransitionFilter transitionFilter,SymbolTable symTab,IOracle oracle);
	}
	
	public List<State> advanceGoldStates(List<State> states,int beam,Instance inst,IOracle oracle){
		for(State s:states)
			if(!s.correct)
				throw new Error("incorrect state?!");
		
		List<State> out=new ArrayList<State>();
		
		if(states.isEmpty())
			return out;
		
		if(mergeStates){
			List<ExpansionItem> candidates=new ArrayList<ExpansionItem>();
			int tieBreaker=0;
			for(State s:states){
//				List<ITransition> correct=DynamicOracleTrainer.getCorrectTransitions(s, inst, transitionFilter, symTab, oracle);
				List<ITransition> correct=getCorrectTransitions(s,inst,oracle);
				int[] trans=iTransitionToIntArray(correct);
				if(trans.length==0)
					throw new Error("this should not happen");
				double[] newScores=scoreTransitions(s,inst,trans);
				for(int i=0;i<newScores.length;++i) //Collect all expansions
					candidates.add(new ExpansionItem(symTab.transition2label[trans[i]],newScores[i],s,tieBreaker++));
			}
			Collections.sort(candidates,Collections.reverseOrder());
			
			HashMap<State,State> seen=new HashMap<State,State>(); //Too keep track of what we've seen
			for(ExpansionItem item:candidates){ //Loop over candidates until we have enough to output
				State s=item.execute();
				if(seen.containsKey(s)){ //Skip identical ones
					continue;
				}
				if(out.size()<beam)
					out.add(s);
				seen.put(s,s);
				if(out.size()==beam) //XXX this is actually bad for sharing correctness (but this kind of only happens with static oracle and merging, so it's kind of a weird thing). Actually, with a dynamic oracle we don't need to share correctness 
					break;
			}
			
		} else {
			MinHeap<ExpansionItem> heap=new MinHeap<ExpansionItem>(new ExpansionItem[beam]);
			int tieBreaker=0;
			for(State s:states){
				List<ITransition> correct=getCorrectTransitions(s, inst, oracle);
				int[] t=iTransitionToIntArray(correct);
				if(t.length==0)
					throw new Error("this should not happen");
				double[] newScores=scoreTransitions(s,inst,t);
				for(int i=0;i<newScores.length;++i){
					if(heap.offer(newScores[i]))
						heap.insert(new ExpansionItem(symTab.transition2label[t[i]],newScores[i],s,tieBreaker++));
				}
			}
			ExpansionItem[] proceed=new ExpansionItem[heap.size()];
			heap.emptyReverseInto(proceed);
			for(int i=0;i<proceed.length;++i)
				out.add(proceed[i].execute());
		}
		
		return out;
	}
	
	private List<ITransition> getCorrectTransitions(State s, Instance inst,	IOracle oracle) {
		if(oracle.isDynamic()){
			return DynamicOracleTrainer.getCorrectTransitions(s, inst, transitionFilter, symTab, oracle);
		} else {
			return Collections.singletonList(symTab.transition2label[inst.transSequence[s.totalTransitions]]);
		}
	}

	static int[] iTransitionToIntArray(List<ITransition> transitions) {
		int[] t=new int[transitions.size()];
		int q=0;
		try {
			for(ITransition it:transitions)
				t[q++]=it.getId();
		} catch(NullPointerException e){
			throw new IllegalArgumentException("transition not defined!!");
		}
		return t;
	}
	
	public List<State> advanceStates(List<State> states,int beam, Instance inst){
		return advanceStates(states,beam,inst,false,null);
	}
	
	public List<State> advanceStates(List<State> states,int beam, Instance inst,boolean augmentLoss, Set<IntPair> goldSwaps){
		if(states.size()==0)
			return Collections.emptyList();
		List<State> out=new ArrayList<State>();
		if(mergeStates){
			List<ExpansionItem> candidates=new ArrayList<ExpansionItem>();
			int tieBreaker=0;
			for(State s:states){
				int[] trans=transitionFilter.getAllowedTransitions(s,inst);
				double[] newScores=scoreTransitions(s,inst,trans);
				for(int i=0;i<newScores.length;++i) //Collect all expansions
					candidates.add(new ExpansionItem(symTab.transition2label[trans[i]],newScores[i],s,tieBreaker++));
			}

			//Now need to go over the list and realize the first beam unique ones
			//Sort candidates descendingly
			Collections.sort(candidates,Collections.reverseOrder());
			
			HashMap<State,State> seen=new HashMap<State,State>(); //Too keep track of what we've seen
			for(ExpansionItem item:candidates){ //Loop over candidates until we have enough to output
				State s=item.execute();
				if(seen.containsKey(s)){ //Skip identical ones
					continue;
				}
				if(out.size()<beam)
					out.add(s);
				seen.put(s,s);
				if(out.size()==beam) //XXX this is actually bad for sharing correctness (but this kind of only happens with static oracle and merging, so it's kind of a weird thing). Actually, with a dynamic oracle we don't need to share correctness 
					break;
			}
		} else {
			//Build a heap of best expansions
			MinHeap<ExpansionItem> heap=new MinHeap<ExpansionItem>(new ExpansionItem[beam]);
			int tieBreaker=0;
			for(State s:states){
				int[] trans=transitionFilter.getAllowedTransitions(s,inst);
//				if(Options.DEBUG){
//					System.err.println("Allowed Transitions: "+Arrays.toString(trans));
//				}
				double[] newScores=scoreTransitions(s,inst,trans);
				if(augmentLoss)
					augmentLoss(s,trans,newScores,inst.head,inst.lbl,goldSwaps);
				for(int i=0;i<newScores.length;++i){
					if(heap.offer(newScores[i]))
						heap.insert(new ExpansionItem(symTab.transition2label[trans[i]],newScores[i],s,tieBreaker++));
				}
			}
			//Then make states out of them
			ExpansionItem[] proceed=new ExpansionItem[heap.size()];
			heap.emptyReverseInto(proceed);
			for(int i=0;i<proceed.length;++i)
				out.add(proceed[i].execute());
		}
		return out;
	}

	private void augmentLoss(State s, int[] trans, double[] newScores,int[] heads,int[] labels, Set<IntPair> goldSwaps) {
		if(!Options.AUG_LOSS)
			return;
		if(!s.hasMinTwoStackElements())
			return;
		int s0=s.s0;
		int s1=s.tail.s0;
		boolean shouldLA=heads[s1]==s0;
		boolean shouldRA=heads[s0]==s1;
		boolean shouldSwap=goldSwaps!=null && goldSwaps.contains(new IntPair(s0,s1));
 
		for(int i=0;i<trans.length;++i){
			ITransition tr=symTab.transition2label[trans[i]];
			if(tr.getTR()==TR.SH)
				continue;
			if(goldSwaps==null && tr.getTR()==TR.SW) //continue if we dont use swap cost
				continue;
			
			if(tr.getTR()==TR.LA){
				if(shouldLA){
					if(((LabeledAbstractTransition) tr).l==labels[s1]){
						//correct label, do nothing
					} else {
						newScores[i]+=LossFunction.ARC_L_FP_LOSS;
					}
				} else {
					newScores[i]+=LossFunction.ARC_FP_LOSS;
				}
			} else if(tr.getTR()==TR.RAS){
				if(shouldRA){
					if(((LabeledAbstractTransition) tr).l==labels[s0]){
						//do nothing
					} else {
						newScores[i]+=LossFunction.ARC_L_FP_LOSS;
					}
				} else {
					newScores[i]+=LossFunction.ARC_FP_LOSS;
				}
			} else if(tr.getTR()==TR.SW){ //This bit increases scores on false positive swaps, false negative swaps will have an increased score on all arc transitions, thereby accounting for the extra cost
				if(!shouldSwap)
					newScores[i]+=LossFunction.SWAP_FP_LOSS;
			}
		}
	}

	@Override
	public State parse(ISentence s) {
		Instance inst=instanceMapping.createTestInstance(s, symTab);
		State state;
		if(beam<=1)
			state=decodeGreedy(inst);
		else
			state=decodeBeam(inst,beam);
		return state;
	}
	
	@Override
	public Pair<State, List<Parse>> parseNBest(ISentence s) {
		Instance inst=instanceMapping.createTestInstance(s, symTab);
		MinHeap<TerminalState> bestFinal=beamUntilAllTerminal(inst,beam);
		TerminalState[] bestSorted=new TerminalState[bestFinal.size()];
		bestFinal.emptyReverseInto(bestSorted);
		State bestState=bestSorted[0].state;
		List<Parse> parses=new ArrayList<Parse>();
		for(TerminalState ts:bestSorted){
			Parse p=ts.toParse(this);
			parses.add(p);
		}
		return new Pair<>(bestState,parses);
	}

	public State decodeBeam(Instance inst, int beamSize) {
		MinHeap<TerminalState> bestFinal=beamUntilAllTerminal(inst,beamSize);
		TerminalState ts=MinHeap.extractMax(bestFinal);
		return ts.state;
	}
	
	private MinHeap<TerminalState> beamUntilAllTerminal(Instance inst,int beamSize){
		MinHeap<TerminalState> bestFinal=new MinHeap<TerminalState>(new TerminalState[beamSize]);		
		List<State> agenda=new ArrayList<State>(beamSize);
		agenda.add(State.getInitialState(inst.head.length,labeled,null));
		while(agenda.size()!=0){
			List<State> expansions=advanceStates(agenda,beamSize,inst);
			filterExpansionsAndReseedAgenda(expansions,agenda,bestFinal);
		}
		return bestFinal;
	}

	public State decodeGreedy(Instance inst) {
		List<State> agenda=Collections.singletonList(State.getInitialState(inst.head.length,labeled,null));
		while(!agenda.get(0).isTerminal())
			agenda=advanceStates(agenda,1,inst);

		return agenda.get(0);
	}

	public State applyTransition(int tr,State s,Instance inst) {
		double[] score=scoreTransitions(s,inst,new int[]{tr});
		return symTab.transition2label[tr].apply(s, score[0]);
	}

	public static State extractBestTerminal(List<State> agenda){
		State best=null;
		for(State s:agenda){
			if(s.isTerminal() && (best==null || s.score()>best.score()))
				best=s;
		}
		return best;
	}

	public static List<State> extractNonTerminal(List<State> newCandidates) {
		List<State> newAgenda=new ArrayList<State>();
		for(State s:newCandidates)
			if(!s.isTerminal())
				newAgenda.add(s);
		return newAgenda;
	}

	public List<Pair<ITransition,Double>> getScoredTransitions(State state,Instance inst) {
		int[] tri=transitionFilter.getAllowedTransitions(state, inst);
		double[] scores=scoreTransitions(state,inst,tri);
		List<Pair<ITransition,Double>> transitions=new ArrayList<Pair<ITransition,Double>>();
		for(int i=0;i<tri.length;++i)
			transitions.add(new Pair<ITransition,Double>(symTab.transition2label[tri[i]],scores[i]));
		return transitions;
	}

	static class TerminalState extends MinHeap.AbstractTieBreakingHeapItem<TerminalState> {

		final State state;
		
		protected TerminalState(int tieBreaker,State state) {
			super(tieBreaker);
			this.state=state;
		}

		public Parse toParse(AbstractDecoder d) {
			Pair<int[],String[]> pair=d.state2tree(this.state);
			Map<String,String> propMap=new HashMap<String,String>();
			propMap.put(Parse.PARSE_PROPERTY_MAP_IMSTRANS_PARSE_SCORE, Double.toString(score()));
			Parse parse=new Parse(pair.getLeft(),pair.getRight(),propMap);
			return parse;
		}

		@Override
		public double score() {
			return state.score();
		}
		
	}
	
	public State decodeGold(ISentence s,int beam,IOracle oracle){
//		Instance inst=instanceMapping.createTestInstance(s, symTab);
		Instance inst=instanceMapping.createTrainInstance(s, symTab, oracle, labeled);
		oracle.initSentence(inst);
		return decodeGold(inst,beam,oracle);
	}
	
	public State decodeGold(Instance inst,int beam,IOracle oracle){
		if(beam>1)
			return decodeGoldBeam(inst,beam,oracle);
		else
			return decodeGoldGreedy(inst,oracle);
	}
	
	protected Pair<State,State[][]> decodeGoldBeamSaveAgendas(Instance inst,int beam,IOracle oracle){
		List<List<State>> stateHistory=new ArrayList<>();
		MinHeap<TerminalState> bestFinal=new MinHeap<TerminalState>(new TerminalState[beam]);
		List<State> agenda=new ArrayList<>();
		agenda.add(State.getInitialState(inst.form.length, labeled, inst.head));
		stateHistory.add(new ArrayList<>(agenda));
		while(!agenda.isEmpty()){
			List<State> advanced=advanceGoldStates(agenda,beam,inst,oracle);
			stateHistory.add(new ArrayList<>(advanced));
			filterExpansionsAndReseedAgenda(advanced,agenda,bestFinal);
		}
		
		State best=MinHeap.extractMax(bestFinal).state;
		
		State[][] hist=new State[best.totalTransitions][];
		for(int i=0;i<hist.length;++i){
			List<State> l=stateHistory.get(i);
			hist[i]=l.toArray(new State[l.size()]);
		}
		
		Pair<State,State[][]> r=new Pair<>(best,hist);
		return r;
	}
	
	protected State decodeGoldBeam(Instance inst,int beam,IOracle oracle){
		MinHeap<TerminalState> bestFinal=new MinHeap<TerminalState>(new TerminalState[beam]);
//		List<State> agenda=Collections.singletonList(State.getInitialState(inst.form.length, labeled, inst.head));
		List<State> agenda=new ArrayList<>();
		agenda.add(State.getInitialState(inst.form.length, labeled, inst.head));
		while(!agenda.isEmpty()){
//			List<State> nextAgenda=new ArrayList<State>();
			List<State> advanced=advanceGoldStates(agenda,beam,inst,oracle);
			filterExpansionsAndReseedAgenda(advanced,agenda,bestFinal);
//			for(State a:advanced){
//				if(a.isTerminal()){
//					if(bestFinal.offer(a.score()))
//						bestFinal.insert(new TerminalState(bestFinal.getNextTieBreaker(),a));
//				} else {
//					if(bestFinal.hasRoomLeft() || bestFinal.peek().score()<a.score())
//						nextAgenda.add(a);
//				}
//			}
//			agenda=nextAgenda;
		}
		TerminalState ts=MinHeap.extractMax(bestFinal);
		return ts.state;
	}
	
//	agenda.clear();
//	for(State s:newItems){
//		if(s.isTerminal()){
//			if(terminalStates.offer(s.score()))
//				terminalStates.insert(new TerminalState(terminalStates.getNextTieBreaker(),s));
//			//else throw it away
//		} else {
//			if(terminalStates.hasRoomLeft() || terminalStates.peek().score()<s.score())
//				agenda.add(s);
//			//else throw it away and don't bother exploring anymore -- it's already worst than the k'th best terminal item, for beam size k --- this is our stopping criterion for terminating the beam search gracefully when we can have derivations of varying lengths
//		}
//	}
	
	public State decodeGoldGreedy(Instance inst,IOracle oracle){
		State s=State.getInitialState(inst.form.length, labeled, inst.head);
		while(!s.isTerminal()){
			List<ITransition> correct=getCorrectTransitions(s,inst,oracle);
			int[] t=iTransitionToIntArray(correct);
			if(t.length==0)
				throw new Error("this should not happen");

			double[] newScores=scoreTransitions(s,inst,t);
			int bestIdx=0;
			for(int i=1;i<newScores.length;++i) //Collect all expansions
				if(newScores[i]>newScores[bestIdx])
					bestIdx=i;
			ITransition tr=correct.get(bestIdx);
			s=tr.apply(s, s.totalScore+newScores[bestIdx]);
		}
		return s;
	}

	
	public static void filterExpansionsAndReseedAgenda(List<State> newItems,List<State> agenda,MinHeap<TerminalState> terminalStates){
		agenda.clear();
		for(State s:newItems){
			if(s.isTerminal()){
				if(terminalStates.offer(s.score()))
					terminalStates.insert(new TerminalState(terminalStates.getNextTieBreaker(),s));
				//else throw it away
			} else {
				if(terminalStates.hasRoomLeft() || terminalStates.peek().score()<s.score())
					agenda.add(s);
				//else throw it away and don't bother exploring anymore -- it's already worst than the k'th best terminal item, for beam size k --- this is our stopping criterion for terminating the beam search gracefully when we can have derivations of varying lengths
			}
		}
	}

	public void overrideBeam(int overRideBeam) {
		this.beam=overRideBeam;
	}

	public State pollBestTerminalExpansion(List<State> agenda, Instance inst) {
		ExpansionItem best=null;
		for(State s:agenda){
			if(s.bufferSize()==0 && s.tail!=null && s.tail.s0==0){ //This corresponds to exactly 2 items on stack and nothing else, only state where terminal state is only one transition away
				int[] trans=transitionFilter.getAllowedTransitions(s,inst);
				double[] newScores=scoreTransitions(s,inst,trans);
				int idx=ArrayFunctions.getHighestIdx(newScores);
				if(best==null || newScores[idx]>best.newTotalScore)
					best=new ExpansionItem(symTab.transition2label[trans[idx]],newScores[idx],s,0);
			}
		}
		if(best==null)
			return null;
		else
			return best.execute();
	}
	
}
