package ims.trans.util;

import java.util.List;

import ims.data.ISentence;
import ims.io.ReaderWriterFactory;
import ims.oracle.IOracle;
import ims.oracle.OracleFactory;
import ims.trans.Options;
import ims.util.ArrayFunctions;
import ims.util.NonProjInfo;

public class AccNPAndSwapInfo {

	public static void main(String[] args) throws Exception{
		Options options=new Options(args);
		IOracle oldSwapOracle=OracleFactory.getStaticOracle(Options.TransitionSystem.ArcSOldSwap);
		IOracle newSwapOracle=OracleFactory.getStaticOracle(Options.TransitionSystem.ArcSNewSwap);
		IOracle optSwapOracle=OracleFactory.getStaticOracle(Options.TransitionSystem.ArcSOptimalSwap);
		Iterable<? extends ISentence> it=ReaderWriterFactory.getReader(options.train, options.format);
		
		int[] npBins=new int[100];           //sentences with n np arcs
		int[] intervalDegBins=new int[100];  //max(interval degree) with Havelkas terminology, gap degree in Kuhlmann and Nivre's terminology
		int[] componentDegBins=new int[100]; //max(component degree) with Havelkas terminology, edge degree in Kuhlmann and Nivre's terminology
		int[] rootDepBins=new int[100];      //sentences with n root deps
		int[] length=new int[1000];
		int[] npLenBins=new int[1000];
		
		int oldSwaps=0;
		int newSwaps=0;
		int optSwaps=0;
		int sentCount=0;
		int accLength=0;
		
		int totalNPArcs=0;
		for(ISentence sen:it){
			++sentCount;
			oldSwaps+=countSwaps(sen,oldSwapOracle);
			newSwaps+=countSwaps(sen,newSwapOracle);
			optSwaps+=countSwaps(sen,optSwapOracle);
			
			int[] heads=sen.getGoldHeads();
			
			NonProjInfo npi=new NonProjInfo(heads);
			int np=npi.totalNonProjArcs;
			npBins[np]++;
			
			rootDepBins[npi.rootDeps]++;
			
			if(np>0){
				int ideg=ArrayFunctions.max(npi.intervalDegree);
				int cdeg=maxCDeg(npi);
				intervalDegBins[ideg]++;
				componentDegBins[cdeg]++;
				
				for(int dep=1;dep<sen.getLength();++dep){
					if(npi.isNonProjectiveEdge[dep]){
						++totalNPArcs;
						int dist=Math.abs(dep-heads[dep]);
						npLenBins[dist]++;
					}
				}
			} else {
				intervalDegBins[0]++;
				componentDegBins[0]++;				
			}
			
			length[sen.getLength()]++;
			accLength+=sen.getLength();
		}
		
		double oRatio=100.0;
		double nRatio=100.0*newSwaps/oldSwaps;
		double pRatio=100.0*optSwaps/oldSwaps;
		
		System.out.println("Total sentences: "+sentCount);
		System.out.printf("Average length: %6.2f\n",(double)accLength/sentCount);
		System.out.printf("Median length:  %6.2f\n",ArrayFunctions.medianFromCountArray(length));
		
		System.out.println();
		System.out.printf("Number of NP arcs:  %d\n",totalNPArcs);
		System.out.printf("Average number of NP arcs per sentence: %5f\n",(double)totalNPArcs/sentCount);
		
		System.out.println();
		System.out.printf("Average NP arc length:  %5f\n",ArrayFunctions.averageFromCountArray(npLenBins));
		System.out.printf("Median NP arc length:  %5f\n",ArrayFunctions.medianFromCountArray(npLenBins));
		
		System.out.println();
				
		System.out.printf("Old swaps: %-6d    (%5.2f%%)\n",oldSwaps,oRatio);
		System.out.printf("New swaps: %-6d    (%5.2f%%)\n",newSwaps,nRatio);
		System.out.printf("Opt swaps: %-6d    (%5.2f%%)\n",optSwaps,pRatio);
		
		System.out.println("\nSentences binned by number of NP arcs\n");
		for(int i=0;i<npBins.length;++i){
			if(npBins[i]==0)
				continue;
			double ratio=100.0*npBins[i]/sentCount;
			System.out.printf("%-2d     %-5d     (%5.2f%%)\n",i,npBins[i],ratio);
		}
		
		System.out.println("\nSentences binned by number of root dependents\n");
		for(int i=0;i<rootDepBins.length;++i){
			if(rootDepBins[i]==0)
				continue;
			double ratio=100.0*rootDepBins[i]/sentCount;
			System.out.printf("%-2d     %-5d     (%5.2f%%)\n",i,rootDepBins[i],ratio);
		}
		
		System.out.println("\nSentences binned by tree interval degree\n");
		for(int i=0;i<intervalDegBins.length;++i){
			if(intervalDegBins[i]==0)
				continue;
			double ratio=100.0*intervalDegBins[i]/sentCount;
			System.out.printf("%-2d     %-5d     (%5.2f%%)\n",i,intervalDegBins[i],ratio);
		}
		
		System.out.println("\nSentences binned by tree component degree\n");
		for(int i=0;i<componentDegBins.length;++i){
			if(componentDegBins[i]==0)
				continue;
			double ratio=100.0*componentDegBins[i]/sentCount;
			System.out.printf("%-2d     %-5d     (%5.2f%%)\n",i,componentDegBins[i],ratio);
		}
		
		System.out.println("\nSentences binned by sentence length\n");
		for(int i=0;i<length.length;++i){
			if(length[i]==0)
				continue;
			double ratio=100.0*length[i]/sentCount;
			System.out.printf("%-4d   %-5d    (%5.2f%%)\n",i,length[i],ratio);
		}
		
		System.out.println("\nNP arcs binned by length\n");
		for(int i=0;i<npLenBins.length;++i){
			if(npLenBins[i]==0)
				continue;
			double ratio=100.0*npLenBins[i]/totalNPArcs;
			System.out.printf("%-3d     %-6d     (%5.2f%%)\n",i,npLenBins[i],ratio);
		}

	}

	private static int maxCDeg(NonProjInfo npi) {
		int max=0;
		for(int i=1;i<npi.componentRoots.length;++i)
			if(npi.componentRoots[i].length>max)
				max=npi.componentRoots[i].length;
		return max;
	}

	private static int countSwaps(ISentence sen, IOracle oracle) {
		int[] heads=sen.getGoldHeads();
		oracle.initSentence(heads);
		List<String> l=oracle.getCanonicalSequence(heads, null, false);
		int r=0;
		for(String s:l)
			if(s.equals("SW"))
				++r;
		return r;
	}
}
