package ims.util.visualtrans;

import ims.oracle.Configuration;
import ims.oracle.Configuration.TR;
import ims.util.DisjointSetZeroToN;
import ims.util.DisjointSetZeroToNRankPathComp;

import java.awt.Color;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.googlecode.whatswrong.NLPInstance;

public class ParseContext {

	public enum ColorCoding { None, OnlyLast, OnlyNonProj, LastAndNonProj };
	
	final WhatsWrongExport wwExport;
	
	final String[] t;
	final Map<String,byte[]> imgCache;
	final Map<String,Configuration> configs;
	TR[] trs;
	
	
	Configuration current;

	double scaleFactor=1.d;
	boolean curved;
	ColorCoding cc;
	
	
	public ParseContext(List<String> tokens,TR[] trs,boolean curved,ColorCoding cc) throws IOException{
		wwExport=new WhatsWrongExport();
		wwExport.setCurved(curved);
		wwExport.setFlagDepsColor(Color.RED);
		tokens.add(0, "<root>");
		this.t=tokens.toArray(new String[tokens.size()]);
		this.imgCache=new HashMap<String,byte[]>();
		this.configs=new HashMap<String,Configuration>();
		this.current=new Configuration(new int[tokens.size()]);
		String sig=current.getTransitionSeqSignature();
		configs.put(sig, current);
		this.curved=curved;
		this.cc=cc;
		this.trs=trs;
	}

	private String[] getColorCodedDeps(Configuration c, ColorCoding cc,	int lastDep) {
		String[] depTypes=new String[c.getPredHeads().length];
		Arrays.fill(depTypes, WhatsWrongExport.DEPTYPE_BLACK);
		if(cc==ColorCoding.None){
			return depTypes;
		}
		if(cc!=ColorCoding.OnlyNonProj && lastDep>0){
			depTypes[lastDep]=WhatsWrongExport.DEPTYPE_RED;
		}
		if(cc==ColorCoding.OnlyLast){
			return depTypes;
		}
		List<Integer> np=getNonProjective(c.getPredHeads());
		for(Integer i:np){
			String color=WhatsWrongExport.DEPTYPE_BLUE;
			if(cc==ColorCoding.LastAndNonProj && i==lastDep)
				color=WhatsWrongExport.DEPTYPE_PURPLE;
			depTypes[i]=color;
		}
		return depTypes;
	}

	private List<Integer> getNonProjective(int[] predHeads) {
		DisjointSetZeroToN components=new DisjointSetZeroToNRankPathComp(predHeads.length);
		for(int i=1;i<predHeads.length;++i)
			if(predHeads[i]>=0)
				components.union(predHeads[i], i);
		List<Integer> r=new ArrayList<Integer>();
		for(int dep=1;dep<predHeads.length;++dep){
			if(predHeads[dep]==-1)
				continue;
			int head=predHeads[dep];
			boolean projective=true;
			int min=Math.min(head, dep);
			int max=Math.max(head, dep);
			for(int btw=min+1;btw<max;++btw){
				if(components.find(btw)!=components.find(head))
					continue;
				int c=btw;
				while(c!=-1){
					if(c==min || c==max)
						break;
					if(predHeads[c]==-1 && (c<min || c>max)){
						projective=false;
						break;
					}
					c=predHeads[c];
				}
			}
			if(!projective)
				r.add(dep);
		}
		return r;
	}

	private String getPNGSig(String stateSig, boolean curved, ColorCoding cc) {
		return stateSig+"+"+(curved?"C":"S")+"+"+cc.toString();
	}

	private byte[] renderConfig(Configuration c,String[] depTypes) throws IOException {
		ByteArrayOutputStream baos=new ByteArrayOutputStream();
		NLPInstance inst=WhatsWrongExport.getNLPInstance(t, c.getPredHeads(), true,depTypes);
		wwExport.exportToPNG(baos, inst, scaleFactor);
		return baos.toByteArray();
	}

	public byte[] getPNG(String stateSig,boolean curved,ColorCoding cc) throws IOException {
		String pngSig=getPNGSig(stateSig,curved,cc);
		byte[] r=imgCache.get(pngSig);
		if(r==null){
			r=applyActionsAndRender(pngSig);
			imgCache.put(pngSig, r);
		}
		return r;
	}

	private byte[] applyActionsAndRender(String sig) throws IOException {
		Configuration c=new Configuration(new int[t.length]);
		int lastDep=-1;
		for(int i=1;i<sig.length();++i){
			char ch=sig.charAt(i);
			if(ch=='+')
				break;
			TR tr=VisualizationServlet.k2TRLookup.get(Character.toString(ch));
			char ch2=sig.charAt(i+1);
			if(ch2=='+'){
				lastDep=getToFlag(tr,c);
			}
			c.applyTR(tr);
		}
		String[] depTypes=getColorCodedDeps(c, cc, lastDep);
		return renderConfig(c,depTypes);
	}

	public void applyTr(TR tr, String lastSig) throws IOException {
		String newSig=lastSig+tr.k;
		Configuration c=configs.get(newSig);
		if(c==null){
			c=configs.get(lastSig).copy();
			c.applyTR(tr);
			String sig=c.getTransitionSeqSignature();
			configs.put(sig, c);
		}
		current=c;
	}

	private int getToFlag(TR tr, Configuration c) {
		int flagDep=-1;
		if(tr==null)
			return flagDep;
		switch(tr){
		case LA:  flagDep=c.peekStack(1); break;
		case RAS: flagDep=c.peekStack(0); break;
		default:
		}
		return flagDep;
	}

	public void setCurrent(String sig) {
		current=configs.get(sig);
	}

	public void setCurved(boolean curved){
		if(this.curved==curved)
			return;
		wwExport.renderer.setCurved(curved);
		this.curved=curved;
	}
	
	public void setScaleFactor(double scaleFactor){
		if(this.scaleFactor==scaleFactor)
			return;
		this.scaleFactor=scaleFactor;
		imgCache.clear();
	}
}
